/*
############################################################################
# Copyright (c) 2010 - 2014 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "links.h"
#include "strsubs.h"
#include "vars.h"
#include "cline.h"
#include "tokens.h"
#include "cdirect.h"

#define OPCODE_MINUS  1
#define OPCODE_PLUS   2
#define OPCODE_MULT   3
#define OPCODE_DIVIDE 4
#define OPCODE_EQUAL  5

#define DSTATE_IDLE 0
#define DSTATE_SKIP 1
#define DSTATE_SHOW 2

LinkT *EmptyLink = 0;
ConInfoT *conlist = 0;
int dlevel = 0;
int skipflag = SKIP_NONE;

void InitDirect(void)
{
    dlevel = 0;
    skipflag = SKIP_NONE;
    EmptyLink = LinkCreate("", sizeof(LinkT));
    AddToConList("CSPIN_FLAG", EmptyLink, CONTYPE_NONE);
}

void AddToConList(char *nstr, LinkT *tokenlist, int type)
{
    ConInfoT *coninfo = (ConInfoT *)LinkCreate(nstr, sizeof(ConInfoT));
    int ival = 0;
    double fval;

    coninfo->str1 = AddToSymbolList(nstr);
    if (coninfo->str1 == 0) coninfo->str1 = "WHOA!!!";
    coninfo->tokenlist = DuplicateTokenList(tokenlist);
    conlist = (ConInfoT *)ListAppend((LinkT *)conlist, (LinkT *)coninfo);
    if (type == CONTYPE_UNKNOWN)
        type = GetConstantTypeList(tokenlist, &fval, &ival);
    coninfo->type = type;
    coninfo->value = ival;
}

ConInfoT *DestroyConList(ConInfoT *conlist)
{
    ConInfoT *coninfo = conlist;

    while (coninfo)
    {
	ListDestroy(coninfo->tokenlist);
	coninfo = coninfo->next;
    }
    ListDestroy((LinkT *)conlist);

    return 0;
}

ConInfoT *AddToMacroList(char *nstr, LinkT *tokenlist, ConInfoT *macrolist)
{
    ConInfoT *coninfo = (ConInfoT *)LinkCreate(nstr, sizeof(ConInfoT));
    coninfo->tokenlist = tokenlist;
    macrolist = (ConInfoT *)ListAppend((LinkT *)macrolist, (LinkT *)coninfo);
    return macrolist;
}

char *GetFloatString(double x)
{
    double y = x;
    double xexp, xmul;
    static char str[100];

    if (x < 0.0) x = -x;

    if (x < 1000000.0 && x > 0.00001)
    {
	sprintf(str, "%lf", y);
    }
    else
    {
	xexp = floor(log10(x));
	xmul = pow(10.0, xexp);
	sprintf(str, "%lfe%d", y/xmul, (int)xexp);
    }

    return str;
}

int GetNumber(char *str, double *pfval, int *pival)
{
    int type = 0;
    char *ptr = FindChar(str, '.');

    *pfval = 0.0;
    *pival = 0;

    if (str[0] == '0' && (str[1] == 'x' || str[1] == 'X'))
    {
	sscanf(str + 2, "%x", pival);
	type = CONTYPE_INT;
    }
    else if (*ptr)
    {
	sscanf(str, "%lf", pfval);
	type = CONTYPE_FLOAT;
    }
    else
    {
	sscanf(str, "%d", pival);
	type = CONTYPE_INT;
    }

    return type;
}

void UpdateValues(int opcode, int type, double fval, int ival, double *fsum, int *isum)
{
    //printf("%s - type = %d, fval = %lf, ival = %d\n",
    //    tokens[i], type, fval, ival);
    if (opcode == OPCODE_MINUS)
    {
        if (type == CONTYPE_FLOAT) *fsum -= fval;
	else if (type == CONTYPE_INT) *isum -= ival;
    }
    else if (opcode == OPCODE_PLUS)
    {
        if (type == CONTYPE_FLOAT) *fsum += fval;
	else if (type == CONTYPE_INT) *isum += ival;
    }
    else if (opcode == OPCODE_MULT)
    {
        if (type == CONTYPE_FLOAT) *fsum *= fval;
	else if (type == CONTYPE_INT) *isum *= ival;
    }
    else if (opcode == OPCODE_DIVIDE)
    {
        if (type == CONTYPE_FLOAT) *fsum /= fval;
	else if (type == CONTYPE_INT) *isum /= ival;
    }
    else if (opcode == OPCODE_EQUAL)
    {
        if (type == CONTYPE_FLOAT) *fsum = fval;
	else if (type == CONTYPE_INT) *isum = ival;
    }
}

static int GetConstantType(char **tokens, int num, double *pfval, int *pival)
{
    int i;
    int type = 0;
    int ival = 0;
    double fval = 0.0;
    int sumtype = 0;
    int isum = 0;
    double fsum = 0.0;
    //char buffer[CLINE_MAX_SIZE];
    //char *tokens[100];
    int opcode = OPCODE_EQUAL;
    //int num = getctokens(str, buffer, tokens, 100, 0) - 1;
    ConInfoT *coninfo;
    char *AllowStrings[] = {"[", "]", "(", ")", 0};

    for (i = 0; i < num; i++)
    {
	//printf(" -%s-", tokens[i]);
	if (tokens[i][0] == '.' || isdigit(tokens[i][0]))
	{
	    type = GetNumber(tokens[i], &fval, &ival);
	    UpdateValues(opcode, type, fval, ival, &fsum, &isum);
	    sumtype = type;
	    opcode = 0;
	}
	else if ((coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, tokens[i])))
	{
	    type = GetConstantTypeList(coninfo->tokenlist, &fval, &ival);
	    UpdateValues(opcode, type, fval, ival, &fsum, &isum);
	    sumtype = type;
	    opcode = 0;
	    if (sumtype == CONTYPE_STRING) break;
	}
	else if (tokens[i][0] == '-')
	    opcode = OPCODE_MINUS;
	else if (tokens[i][0] == '+')
	    opcode = OPCODE_PLUS;
	else if (tokens[i][0] == '*')
	    opcode = OPCODE_MULT;
	else if (tokens[i][0] == '/')
	    opcode = OPCODE_DIVIDE;
	else if (SearchList(tokens[i], AllowStrings, SEARCH_EXACT) < 0)
	{
	    isum = 0;
	    fsum = 0.0;
	    sumtype = CONTYPE_STRING;
	    break;
	}
    }
    *pfval = fsum;
    *pival = isum;
    return sumtype;
}

int GetConstantTypeList(LinkT *tokenlist, double *fval, int *ival)
{
    char *tokens[200];
    int num = ConvertTokenListToArray(tokenlist, tokens, 200);
    return GetConstantType(tokens, num, fval, ival);
}

int EvaluateNumber(char **tokens, int num)
{
    int ival;
    double fval;
    int type = GetConstantType(tokens, num, &fval, &ival);

    if (type == CONTYPE_FLOAT) ival = fval;

    return ival;
}

void PrintConstants(void)
{
    int i = 0;
    int j;
    int type;
    double fval;
    int ival;
    LinkT *list;
    ConInfoT *coninfo = conlist;

    printf("Constants List\n");
    while (coninfo)
    {
        printf("%3d: %s", i++, coninfo->str);
	if (strcmp(coninfo->str, coninfo->str1))
	    printf(" (%s)", coninfo->str1);
	list = coninfo->tokenlist;
	printf(" \"");
	for (j = 0; list; j = 1)
	{
	    if (j) printf(" ");
	    printf("%s", list->str);
	    list = list->next;
	}
	printf("\"");
        type = GetConstantTypeList(coninfo->tokenlist, &fval, &ival);
	if (type == CONTYPE_FLOAT)
	    printf(" %s", GetFloatString(fval));
	else if (type == CONTYPE_INT)
	    printf(" %d", ival);
	printf(", %d", coninfo->type);
	printf("\n");
	coninfo = coninfo->next;
    }
}

void PrintMacroError(char *str)
{
    PrintFileLine();
    printf(" - Error in macro - %s\n", str);
    CspinExit(1);
}

void PrintMacroList(ConInfoT *macrolist)
{
    LinkT *link;

    printf("PrintMacroList\n");
    while (macrolist)
    {
	printf("%s:", macrolist->str);
	link = macrolist->tokenlist;
	while (link)
	{
	    printf(" %s", link->str);
	    link = link->next;
	}
	printf("\n");
	macrolist = macrolist->next;
    }
}

LinkT *ExpandLine(LinkT *tokenlist)
{
    LinkT *head = tokenlist;
    LinkT *list = tokenlist;
    LinkT *prev = 0;
    LinkT *link;
    LinkT *next;
    LinkT *tail;
    char *str;
    ConInfoT *coninfo;
    int loopcount = 0;
    int parencount;

    while (list)
    {
	if (++loopcount >= 1000)
	{
	    PrintFileLine();
	    printf(" - Infinite loop in macro expansion\n");
	    CspinExit(1);
	}
	str = list->str;
	coninfo = (ConInfoT *)ListFindLinkString((LinkT *)conlist, str);

	if (coninfo && coninfo->type == CONTYPE_STRING)
	{
	    link = coninfo->tokenlist;
	    if (link == 0 || (link->str[0] == 0 && link->next == 0))
	    {
	        next = list->next;

		if (prev)
		    prev->next = next;
		else
		    head = next;

	    }
	    else
	    {
	        next = DuplicateTokenList(link);

		if (prev)
		    prev->next = next;
		else
		    head = next;

		tail = ListTail(next);
		tail->next = list->next;
	    }
	    LinkDestroy(list);
	    list = next;
	}
	else if (coninfo && coninfo->type == CONTYPE_MACRO)
	{
	    ConInfoT *macrolist = 0;
	    LinkT *inlist;
	    LinkT *maclist;
	    LinkT *inprev;

	    link = coninfo->tokenlist;
	    // Extract the parameters
	    inlist = list->next;
	    maclist = coninfo->tokenlist;
	    if (inlist->str[0] != '(') PrintMacroError("Expected \"(\"");
	    if (maclist->str[0] != '(') PrintMacroError("Expected \"(\"");
	    while (1)
	    {
	        inlist = inlist->next;
	        maclist= maclist->next;
	        if (!inlist || !maclist) PrintMacroError("Encountered end of line");
		//printf("in = %s, mac = %s\n", inlist->str, maclist->str);
	        macrolist = AddToMacroList(maclist->str, inlist, macrolist);
	        maclist= maclist->next;
	        if (!maclist) PrintMacroError("Encountered end of line");
		if (maclist->str[0] != ',' && maclist->str[0] != ')') PrintMacroError("Unexpected character");
		inprev = 0;
		parencount = 0;
		while (1)
		{
		    if (inlist->str[0] == '(' || inlist->str[0] == '[')
		        parencount++;
		    else if (inlist->str[0] == ')' || inlist->str[0] == ']')
		        parencount--;
		    if (parencount <= 0 && inlist->str[0] == ',') break;
		    if (parencount <  0 && inlist->str[0] == ')') break;
		    //printf("inlist = %s\n", inlist->str);
		    inprev = inlist;
	            inlist = inlist->next;
	            if (!inlist) PrintMacroError("Encountered end of line");
		}
		if (inprev == 0)
		    macrolist->tokenlist = 0;
		else
		{
		    //printf("Breaking the chain at %s\n", inprev->str);
		    inprev->next = 0;
		}
		//printf("in = %s, mac = %s\n", inlist->str, maclist->str);
		if (strcmp(inlist->str, maclist->str) != 0) PrintMacroError("Number of parameters doesn't match");
		if (inlist->str[0] == ')') break;
	    }
	    //PrintMacroList(macrolist);
	    maclist = maclist->next;
            if (!maclist) PrintMacroError("Encountered end of line");
	    //Expand the rest of the macro and copy to the input line
	    list = 0;
	    inprev = prev;
	    while (maclist)
	    {
		// Check macrolist for token
		coninfo = (ConInfoT *)ListFindLinkString((LinkT *)macrolist, maclist->str);
	        if (coninfo)
	        {
	            link = coninfo->tokenlist;

		    if (link && link->str[0])
	            {
	                link = DuplicateTokenList(link);

                        if (list == 0) list = link;
		        if (inprev)
		            inprev->next = link;
		        else
		            head = link;
		        link = ListTail(link);
	            }
		}
		else
		{
		    link = LinkCreate(maclist->str, sizeof(LinkT));
                    if (list == 0) list = link;
		    if (inprev)
		        inprev->next = link;
		    else
		        head = link;
		}
		inprev = link;
		maclist = maclist->next;
	    }

            inlist = inlist->next;
	    if (inprev)
	        inprev->next = inlist;
	    else
	        head = inlist;
#if 0
	    list = inlistt;
	    prev = inprev;
#else
            if (list == 0)
	    {
		list = inlist;
		prev = inprev;
	    }
	    else if (prev)
	        prev->next = list;
#endif

	    DestroyConList(macrolist);
	}
	else
	{
	    prev = list;
	    list = list->next;
	}
    }
    return head;
}

int ProcessDirective(LinkT *tokenlist)
{
    int i;
    int errflag = 0;
    char *buffer = "";
    char *buffer1 = "";
    char *buffer2 = "";
    LinkT *tokenlist0 = tokenlist;
    LinkT *tokenlist1 = EmptyLink;
    static int dstate[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    if (tokenlist == 0) return skipflag;

    buffer = tokenlist->str;
    if (strcmp(buffer, "#") != 0) return skipflag;

    tokenlist = tokenlist->next;
    if (tokenlist == 0)
    {
	PrintFileLine();
	printf(" - Empty preprocessor directive\n");
	CspinExit(1);
    }

    buffer = tokenlist->str;
    tokenlist = tokenlist->next;
    if (tokenlist)
    {
	if (tokenlist->str[0] != '\n') buffer1 = tokenlist->str;
        tokenlist1 = tokenlist;
	tokenlist = tokenlist->next;
	if (tokenlist && tokenlist->str[0] != '\n') buffer2 = tokenlist->str;
	else tokenlist = EmptyLink;
    }

    //printf("ProcessDirective: -%s- -%s- -%s-\n", buffer, buffer1, buffer2);

    if (strcmp(buffer, "define") == 0 || strcmp(buffer, "macro") == 0)
    {
	int type = CONTYPE_UNKNOWN;
	if (buffer[0] == 'm') type = CONTYPE_MACRO;
	if (skipflag != SKIP_MODE)
	{
            LinkT *prev = 0;
            LinkT *list = tokenlist;

	    if (*buffer1 == 0)
	    {
	        PrintFileLine();
	        printf(" - #define has no parameters\n");
	        CspinExit(1);
	    }
	    // Remove newline from tokenlist
	    while (list)
	    {
		if (list->str[0] == '\n')
		{
		    if (prev)
		        prev->next = 0;
		    else
		        tokenlist = EmptyLink;
		    break;
		}
		prev = list;
		list = list->next;
	    }
	    AddToConList(buffer1, tokenlist, type);
	}
	return skipflag;
    }

    if (skipflag != SKIP_MODE) skipflag = SKIP_LINE;

    if (strncmp(buffer, "pragma", 7) == 0) return skipflag;

    if (strncmp(buffer, "include", 8) == 0)
    {
	if (skipflag != SKIP_MODE)
	    ProcessIncludeDirective(tokenlist0);
        return skipflag;
    }

    // Below this point we will set skipflag to either SKIP_LINE or SKIP_MODE

    // Check if we need to bump the level
    if (dstate[dlevel] != DSTATE_IDLE && strncmp(buffer, "if", 2) == 0)
    {
	//printf("Increase dlevel to %d\n", dlevel + 1);
	if (dlevel >= 9)
	{
	    PrintFileLine();
	    printf(" - Too many nested directives\n");
	    CspinExit(1);
	}
	dstate[++dlevel] = DSTATE_IDLE;
    }

    if (dstate[dlevel] == DSTATE_IDLE)
    {
	if (strcmp(buffer, "if") == 0)
	{
            int ival;
            double fval;
            int type = GetConstantTypeList(tokenlist1, &fval, &ival);
	    if (type == CONTYPE_FLOAT) ival = (int)fval;
            if (ival)
	        dstate[dlevel] = DSTATE_SHOW;
            else
	        dstate[dlevel] = DSTATE_SKIP;
#if 0
	    else
	        errflag = 1;
#endif
	}
	else if (strcmp(buffer, "ifdef") == 0)
	{
            if (ListFindLinkString((LinkT *)conlist, buffer1))
	        dstate[dlevel] = DSTATE_SHOW;
	    else
	        dstate[dlevel] = DSTATE_SKIP;
	}
	else if (strcmp(buffer, "ifndef") == 0)
	{
            if (ListFindLinkString((LinkT *)conlist, buffer1))
	        dstate[dlevel] = DSTATE_SKIP;
	    else
	        dstate[dlevel] = DSTATE_SHOW;
	}
	else errflag = 1;
    }
    else if (dstate[dlevel] == DSTATE_SHOW)
    {
	if (strcmp(buffer, "else") == 0)
	{
	    dstate[dlevel] = DSTATE_SKIP;
	}
	else if (strcmp(buffer, "endif") == 0)
	{
	    dstate[dlevel] = DSTATE_IDLE;
	    if (dlevel > 0) dlevel--;
	}
	else errflag = 1;
    }
    else if (dstate[dlevel] == DSTATE_SKIP)
    {
	if (strcmp(buffer, "else") == 0)
	{
	    dstate[dlevel] = DSTATE_SHOW;
	}
	else if (strcmp(buffer, "endif") == 0)
	{
	    dstate[dlevel] = DSTATE_IDLE;
	    if (dlevel > 0) dlevel--;
	}
	else errflag = 1;
    }
    else errflag = 1;

    if (errflag)
    {
        PrintFileLine();
        printf(" - Unexpected directive - %s %s\n", buffer, buffer1);
    }

    // Determine the skipflag
    skipflag = SKIP_LINE;
    for (i = 0; i <= dlevel; i++)
    {
	if (dstate[i] == DSTATE_SKIP)
	{
	    skipflag = SKIP_MODE;
	    break;
	}
    }

    return skipflag;
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
