/*
############################################################################
# Copyright (c) 2010 - 2014 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <string.h>
#include "links.h"
#include "tokens.h"
#include "spinit.h"
#include "vars.h"

extern int methodptr;
extern FILE *outfile;
extern LinkT *FuncPtrList;
extern VarInfoT *varlist;

void AddToFuncPtrList(char *str, int numparms)
{
    LinkT *link = FuncPtrList;

    // Check if already on the list
    while (link)
    {
        if (!strcmp(link->str, str))
            return;
        link = link->next;
    }

    link = AddToken(str, &FuncPtrList);
    link->id = numparms;
}

char *GetFuncPtrName(char *str)
{
    int i;
    static char tempstr[100];

    strcpy(tempstr, "_fp_");

    for (i = 4; i < 99; i++)
    {
        if (*str == '.')
            tempstr[i] = '_';
        else
            tempstr[i] = *str;
        str++;
    }
    tempstr[i] = 0;
    return tempstr;
}

void PrintFuncPtrInit(void)
{
    int i, num;
    LinkT *link = FuncPtrList;


    while (link)
    {
        num = link->id;
        fprintf(outfile, "  if mp.SetMethodPtrEx(@%s)\n",
            GetFuncPtrName(link->str));
        fprintf(outfile, "    %s", link->str);
        if (num)
        {
            fprintf(outfile, "(");
            for (i = 1; i < num; i++)
                fprintf(outfile, "0, ");
            fprintf(outfile, "0)");
        }
        fprintf(outfile, "\n");
        link = link->next;
    }
}

void PrintFuncPtrDat(void)
{
    LinkT *link = FuncPtrList;

    while (link)
    {
        fprintf(outfile, "  %s long 0[2]\n", GetFuncPtrName(link->str));
        link = link->next;
    }
}

int ConvertFuncPtrToSpin(int i, char **tokens1, int num1, LinkT **ptokenlistout, int clinetype)
{
    int num3;
    int i0 = i;
    VarInfoT *varinfo = FindVariable(varlist, tokens1[i]);
    LinkT *link = ListTail(*ptokenlistout);

    if (!varinfo)
    {
        AddToken(tokens1[i], ptokenlistout);
        return i;
    }

    if (link && !strcmp(link->str, "("))
        *ptokenlistout = ListRemoveLink(*ptokenlistout, link);

    /* Locate closing paren */
    for (; i < num1; i++)
    {
        if (tokens1[i][0] == ')')
           break;
    }
    num3 = i;

    if (i+2 < num1 && tokens1[i+2][0] == ')')
    {
        AddToken("mp.CallMethod0", ptokenlistout);
        AddToken("(", ptokenlistout);
        i += 2;
    }
    else
    {
        int num2;
        char tempstr[24];
        int argcount = SpinitCountParms(i+1, tokens1, num1, &num2);
        sprintf(tempstr, "mp.CallMethod%d", argcount+1);
        AddToken(tempstr, ptokenlistout);
        ConvertStatementToSpin(i+1, tokens1, num2, ptokenlistout, clinetype);
        i = num2;
        AddToken(",", ptokenlistout);
    }

    AddToken(varinfo->str1, ptokenlistout);
    ConvertStatementToSpin(i0+1, tokens1, num3, ptokenlistout, clinetype);
    AddToken(")", ptokenlistout);

    return i;
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
